###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         plugins/                                                            #
#                 oneanalysis.py                                              #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import threading
import traceback
import sys
import logging
import peekaboo.queuing
from peekaboo.ruleset import RuleResult
from peekaboo.exceptions import CuckooReportPendingException
from peekaboo.toolbox.sampletools import ConnectionMap


logger = logging.getLogger(__name__)


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]
    return getinstance


@singleton
class OneAnalysis(object):
    """
    @author: Felix Bauer
    """
    __in_use = threading.Lock()
    
    def already_in_progress(self, s):
        with self.__in_use:
            logger.debug("enter already_in_progress")
            tb = traceback.extract_stack()
            tb = tb[-1]
            position = "%s:%s" % (tb[2], tb[1])

            if s.has_attr('pending'):
                s.set_attr('pending', False)
                return RuleResult(position,
                                  result=s.get_result(),
                                  reason='Datei wird jetzt Analysiert',
                                  further_analysis=True)

            l = []
            for sample in ConnectionMap.get_samples_by_sha256(s.sha256sum):
                if sample != s:
                    if not sample.has_attr('pending') or not sample.get_attr('pending') is True:
                        l.append(sample)

            if len(l) == 0:
                s.set_attr("pending", False)
                logger.debug("no second analysis present")
                return RuleResult(position,
                                  result=s.get_result(),
                                  reason='Datei wird jetzt Analysiert',
                                  further_analysis=True)

            logger.debug("there is another same sample")
            logger.debug("I'll be off until needed")
            s.set_attr("pending", True)
            # stop worker
            sys.stdout.flush()
            logger.debug("leave already_in_progress")
            raise CuckooReportPendingException()

    def queue_identical_samples(self, s):
        with self.__in_use:
            logger.debug("queueing identical samples")
            for sample in ConnectionMap.get_samples_by_sha256(s.sha256sum):
                pending = sample.get_attr('pending')
                if pending:
                    sample.set_attr('pending', False)
                    peekaboo.queuing.JobQueue.submit(sample, self.__class__)



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# daemon.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from threading import Thread
import SocketServer
import os
import sys
import grp
import pwd
import stat
import logging
from argparse import ArgumentParser
from sdnotify import SystemdNotifier
from twisted.internet import reactor
from peekaboo import _owl, __version__
from peekaboo.config import parse_config, get_config
from peekaboo.db import PeekabooDatabase
from peekaboo.toolbox.cuckoo import CuckooServer
from peekaboo.toolbox.sampletools import ConnectionMap
from peekaboo.queuing import JobQueue, create_workers
from peekaboo.sample import make_sample


logger = logging.getLogger(__name__)


class PeekabooStreamServer(SocketServer.ThreadingUnixStreamServer):
    """
    Asynchronous server.

    @author: Sebastian Deiss
    """
    def __init__(self, server_address, request_handler_cls, bind_and_activate=True):
        self.config = get_config()
        create_workers(self.config.worker_count)
        # We can only accept 2 * worker_count connections.
        self.request_queue_size = self.config.worker_count * 2
        self.allow_reuse_address = True
        SocketServer.ThreadingUnixStreamServer.__init__(self, server_address,
                                                        request_handler_cls,
                                                        bind_and_activate=bind_and_activate)

    def shutdown_request(self, request):
        """ Keep the connection alive until Cuckoo reports back, so the results can be send to the client. """
        # TODO: Find a better solution.
        pass

    def server_close(self):
        os.remove(self.config.sock_file)
        return SocketServer.ThreadingUnixStreamServer.server_close(self)


class PeekabooStreamRequestHandler(SocketServer.StreamRequestHandler):
    def handle(self):
        """
        Handles an run_analysis request. The path of the directory / file to analyse must
        be written to the corresponding socket.
        The maximum buffer size is 1024 bytes.
        """
        self.request.sendall('Hallo das ist Peekaboo\n\n')
        path = self.request.recv(1024).rstrip()
        logger.info("Got run_analysis request for %s" % path)

        if not os.path.exists(path):
            self.request.sendall(
                'FEHLER: Pfad existiert nicht oder Zugriff verweigert.'
            )
            logger.error("ERROR: Path does not exist or no permission to access it.")
        else:
            for_analysis = []
            if os.path.isfile(path):
                sample = make_sample(path, self.request)
                if sample:
                    for_analysis.append(sample)
            else:
                # walk recursively through entries in the given directory.
                for dirname, __, filenames in os.walk(path):
                    for filename in filenames:
                        logger.debug("Found file %s" % filename)
                        f = os.path.join(dirname, filename)
                        sample = make_sample(f, self.request)
                        if sample:
                            for_analysis.append(sample)

            # introduced after an issue where results were reported
            # before all files could be added.
            for sample in for_analysis:
                ConnectionMap.add(self.request, sample)
                JobQueue.submit(sample, self.__class__)

    # TODO: do cleanup work here in finish()


def run():
    """ Runs the daemon. """
    arg_parser = ArgumentParser()
    arg_parser.add_argument(
        '-c', '--config',
        action='store',
        required=False,
        default=os.path.join('./peekaboo.conf'),
        help='The configuration file for Peekaboo.'
    )
    arg_parser.add_argument(
        '-d', '--debug',
        action='store_true',
        required=False,
        default=False,
        help="Run Peekaboo in debug mode regardless of what's specified in the configuration."
    )
    arg_parser.add_argument(
        '-D', '--daemon',
        action='store_true',
        required=False,
        default=False,
        help='Run Peekaboo in daemon mode (suppresses the logo to be written to STDOUT).'
    )
    args = arg_parser.parse_args()

    if not args.daemon:
        print(_owl)
    else:
        print('Starting Peekaboo %s.' % __version__)

    # read configuration
    config = parse_config(args.config)

    # Check if CLI arguments override the configuration
    if args.debug:
        config.change_log_level('DEBUG')

    # Log the configuration options if we are in debug mode
    if config.log_level == logging.DEBUG:
        logger.debug(config.__str__())

    # establish a connection to the database
    try:
        db_con = PeekabooDatabase(config.db_url)
        config.add_db_con(db_con)
    except Exception as e:
        logger.critical('Failed to establish a connection to the database.')
        logger.exception(e)
        sys.exit(1)

    # Import debug module if we are in debug mode
    if config.use_debug_module:
        from peekaboo.debug import peekaboo_debugger
        peekaboo_debugger()

    if os.getuid() == 0:
        logger.warning('Peekaboo should not run as root.')
        # drop privileges to user
        os.setgid(grp.getgrnam(config.group)[2])
        os.setuid(pwd.getpwnam(config.user)[2])
        # set $HOME to the users home directory
        # (VirtualBox must access the configs)
        os.environ['HOME'] = pwd.getpwnam(config.user)[5]
        logger.info("Dropped privileges to user %s and group %s"
                    % (config.user, config.group))
        logger.debug('$HOME is ' + os.environ['HOME'])

    # write PID file
    pid = str(os.getpid())
    with open(config.pid_file, "w") as pidfile:
        pidfile.write("%s\n" % pid)

    systemd = SystemdNotifier()
    server = PeekabooStreamServer(config.sock_file, PeekabooStreamRequestHandler)
    runner = Thread(target=server.serve_forever)
    runner.daemon = True

    try:
        runner.start()
        logger.info('Peekaboo server is listening on %s' % server.server_address)

        os.chmod(config.sock_file, stat.S_IWOTH | stat.S_IREAD |
                                   stat.S_IWRITE | stat.S_IRGRP |
                                   stat.S_IWGRP | stat.S_IWOTH)

        # Run Cuckoo sandbox, parse log output, and report back of Peekaboo.
        # If this dies Peekaboo dies, since this is the main thread.
        srv = CuckooServer()
        reactor.spawnProcess(srv, config.interpreter, [config.interpreter, '-u',
                                                       config.cuckoo_exec])
        systemd.notify("READY=1")
        reactor.run()
    except Exception as e:
        logger.exception(e)
    finally:
        server.shutdown()


if __name__ == '__main__':
    run()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         files.py                                                            #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import logging
import subprocess
import mimetypes
import magic
from peekaboo.config import get_config


logger = logging.getLogger(__name__)


def chown2me():
    """ Acquire ownership of all directories under /tmp with the prefix "amavis-". """
    # TODO: Find a better solution to acquire ownership and only for the directory currently in usse.
    logger.debug('Invoking chown2me...')
    config = get_config()
    proc = subprocess.Popen(config.chown2me_exec,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    proc.wait()
    if proc.returncode != 0:
        logger.error('chown2me exited with code %d' % proc.returncode)


def guess_mime_type_from_filename(file_path):
    """ Guess the type of a file based on its filename or URL. """
    if not mimetypes.inited:
        mimetypes.init()
        mimetypes.add_type('application/javascript', '.jse')

    mt = mimetypes.guess_type(file_path)[0]
    if mt:
        return mt


def guess_mime_type_from_file_contents(file_path):
    """  Get type from file magic bytes. """
    mt = magic.from_file(file_path, mime=True)
    if mt:
        return mt



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         __init__.py                                                         #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from enum import Enum, unique


@unique
class Result(Enum):
    """
    @author: Felix Bauer
    """
    inProgress = 0
    unchecked = 1
    unknown = 2
    ignored = 3
    checked = 4
    good = 5
    bad = 6

    @staticmethod
    def from_string(result_str):
        for i in Result:
            if i.name == result_str:
                return i
        raise ValueError('%s: Element not found' % result_str)

    def __gt__(self, other):
        return self.value >= other.value

    def __lt__(self, other):
        return other.value >= self.value


class RuleResult:
    """
    @author: Felix Bauer
    """
    def __init__(self, rule,
                 result=Result.unknown,
                 reason='regel ohne Ergebnis',
                 further_analysis=True):
        self.rule = None
        self.result = Result.unchecked
        self.rule = rule
        self.result = result
        self.reason = reason
        self.further_analysis = further_analysis

    def __str__(self):
        return ("Ergebnis \"%s\" der Regel %s - %s, Analyse wird fortgesetzt: %s."\
                                                        % (self.result.name,
                                                           self.rule,
                                                           self.reason,
                                                           'Ja' if self.further_analysis else 'Nein'))

    __repr__ = __str__



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# debug.py                                                                    #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


# DO NOT USE IN PRODUCTION!!
#
# import this to your project and connect to debug2.sock
#  socket $(pwd)/debug2.sock
# execute (by line) python code
# redirects all input/output to socket connection
#
#
# import threading; print threading.activeCount()
#
# import pjobs
# for W in pjobs.Workers.w: print W.isAlive()
# print pjobs.Workers.q.qsize()
# print pjobs.Jobs.jobs


import socket
import os
import fileinput
import sys
import threading
import logging


logger = logging.getLogger(__name__)


def debug():
    sockfile = os.path.abspath('./debug.sock')
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        # attempt to remove an earlier socket file
        os.remove(sockfile)
    except OSError:
        # doesn't matter :-)
        pass

    s.bind(sockfile)
    s.listen(5)

    logger.debug('Peekaboo debugging socket %s created.' % sockfile)
    logger.debug('DO NOT USE THIS MODULE IN PRODUCTION!')
    logger.debug('You may now connect to the socket with:')
    logger.debug('    socket %s' % sockfile)
    logger.debug('Once connected to the socket, you can execute Python code.')

    while True:
        conn, __ = s.accept()

        stdin = sys.stdin
        stdout = sys.stdout
        stderr = sys.stderr

        # sys.stdin  = conn.fileno()
        # sys.stdout = conn.fileno()
        # sys.stderr = conn.fileno()

        os.dup2(conn.fileno(), 0)
        os.dup2(conn.fileno(), 1)
        os.dup2(conn.fileno(), 2)

        sys.stdout.write('> ')
        for line in fileinput.input():
            try:
                exec line
            except Exception as e:
                logger.exception(e)
            sys.stdout.write('> ')

        sys.stdin = stdin
        sys.stdout = stdout
        sys.stderr = stderr


def peekaboo_debugger():
    threading.Thread(target=debug).start()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# db.py                                                                       #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from datetime import datetime
from sqlalchemy import Column, Integer, String, Text, DateTime, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session, relationship
from sqlalchemy.exc import SQLAlchemyError
from peekaboo import __version__
from peekaboo.ruleset import Result, RuleResult
from peekaboo.exceptions import PeekabooDatabaseError
import threading
import logging


logger = logging.getLogger(__name__)
Base = declarative_base()


#
# Database schema definition.
##############################################################################


class PeekabooMetadata(Base):
    """
    Definition of the _meta table for Peekaboo.

    @author: Sebastian Deiss
    """
    __tablename__ = '_meta'
    peekaboo_version = Column(String(10), nullable=False,
                              primary_key=True)
    db_schema_version = Column(Integer, nullable=False,
                               primary_key=True,
                               autoincrement=False)
    cuckoo_version = Column(String(10), nullable=False,
                            primary_key=True)

    def __str__(self):
        return (
            '<PeekabooMetadata(peekaboo_version="%s", db_schema_version="%s", '
            'cuckoo_version="%s")>'
            % (self.peekaboo_version,
               self.db_schema_version,
               self.cuckoo_version)
        )

    __repr__ = __str__


class AnalysisResult(Base):
    """
    Definition of the analysis_result table.

    @author: Sebastian Deiss
    """
    __tablename__ = 'analysis_result_v2'

    id = Column(Integer, primary_key=True)
    name = Column(String(255), nullable=False)

    def __str__(self):
        return '<AnalysisResult(name="%s")>' % self.name

    __repr__ = __str__


class SampleInfo(Base):
    """
    Definition of the sample_info table.

    @author: Sebastian Deiss
    """
    __tablename__ = 'sample_info_v2'

    id = Column(Integer, primary_key=True)
    sha256sum = Column(String(64), nullable=False)
    file_extension = Column(String(16), nullable=True)
    result_id = Column(Integer, ForeignKey('analysis_result_v2.id'),
                       nullable=False)
    result = relationship("AnalysisResult")
    reason = Column(Text, nullable=True)

    def set_result(self, result):
        self.__result = result

    def get_result(self):
        return self.__result

    def __str__(self):
        return ('<SampleInfo(sample_sha256_hash="%s", file_extension="%s", '
                'reason="%s")>'
                % (self.sha256sum,
                   self.file_extension,
                   self.reason))

    __repr__ = __str__


class AnalysisJournal(Base):
    """
    Definition of the analysis_jobs table.

    @author: Sebastian Deiss
    """
    __tablename__ = 'analysis_jobs_v2'

    id = Column(Integer, primary_key=True)
    job_hash = Column(String(255), nullable=False)
    filename = Column(String(255), nullable=False)
    analyses_time = Column(DateTime, nullable=False)
    sample_id = Column(Integer, ForeignKey('sample_info_v2.id'),
                       nullable=False)
    sample = relationship('SampleInfo')

    def __str__(self):
        return (
            '<AnalysisJournal(job_hash="%s", filename="%s", analysis_time="%s")>'
            % (self.job_hash,
               self.filename,
               self.analyses_time.strftime("%Y%m%dT%H%M%S"))
        )

    __repr__ = __str__


#
# End of database schema definition.
##############################################################################


class PeekabooDatabase(object):
    """
    Peekaboo's database.

    @author: Sebastian Deiss
    """
    def __init__(self, db_url):
        """
        Initialize the Peekaboo database handler.

        :param db_url: An RFC 1738 URL that points to the database.
        """
        self.__engine = create_engine(db_url, pool_recycle=1)
        session_factory = sessionmaker(bind=self.__engine)
        self.__Session = scoped_session(session_factory)
        self.__lock = threading.RLock()
        if not self.__engine.dialect.has_table(self.__engine, '_meta'):
            self._init_db()
            logger.debug('Database schema created.')
        else:
            self.clear_in_progress()

    def analysis2db(self, sample):
        """
        Save an analysis task to the analysis journal in the database.

        :param sample: The sample object for this analysis task.
        """
        with self.__lock:
            session = self.__Session()
            analysis = AnalysisJournal()
            analysis.job_hash = sample.get_job_hash()
            analysis.filename = sample.get_filename()
            analysis.analyses_time = datetime.strptime(sample.analyses_time,
                                                       "%Y%m%dT%H%M%S")
            analysis_result = PeekabooDatabase.__get_or_create(
                session,
                AnalysisResult,
                name=sample.get_result().name
            )
            # NOTE: We cannot determine if a known sample is inProgress again.
            s = PeekabooDatabase.__get(
                session,
                SampleInfo,
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension,
            )
            if s is None:
                s = PeekabooDatabase.__create(
                    SampleInfo,
                    sha256sum=sample.sha256sum,
                    file_extension=sample.file_extension,
                    result=analysis_result
                )
            analysis.sample = s
            session.add(analysis)
            try:
                session.commit()
            except SQLAlchemyError as e:
                session.rollback()
                raise PeekabooDatabaseError(
                    'Failed to add analysis task to the database: %s' % e
                )
            finally:
                session.close()

    def sample_info_update(self, sample):
        """
        Update sample information.

        :param sample: The sample object containing the info to update
        """
        with self.__lock:
            session = self.__Session()
            sample_info = PeekabooDatabase.__get(
                session,
                SampleInfo,
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension
            )
            if sample_info is not None:
                sample_info.result = PeekabooDatabase.__get_or_create(
                    session,
                    AnalysisResult,
                    name=sample.get_result().name
                )
                sample_info.reason = sample.reason
                try:
                    session.commit()
                    logger.debug(
                        'Updated sample info in the database for sample %s.' % sample
                    )
                except SQLAlchemyError as e:
                    session.rollback()
                    raise PeekabooDatabaseError(
                        'Failed to update info for sample %s in the database: %s'
                        % (sample, e)
                    )
                finally:
                    session.close()
            else:
                raise PeekabooDatabaseError(
                    'No info found in the database for sample %s' % sample
                )

    def sample_info_exists(self, sample):
        """
        Check if a log for a given sample exists.

        :param sample: The Sample object to check.
        """
        with self.__lock:
            return self.__exists(
                SampleInfo,
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension
            )

    def sample_info_fetch(self, sample):
        with self.__lock:
            session = self.__Session()
            sample_info = PeekabooDatabase.__get(
                session,
                SampleInfo,
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension
            )
            if sample_info:
                sample_info.set_result(Result.from_string(sample_info.result.name))
            session.close()
        return sample_info

    def fetch_rule_result(self, sample):
        """
        Gets the sample information from the database as a RuleResult object.

        :param sample: The Sample object to get the rule result from.
        :return: Returns a RuleResult object containing the sample information.
        """
        with self.__lock:
            session = self.__Session()
            sample_info = PeekabooDatabase.__get(
                session,
                SampleInfo,
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension
            )
            if sample_info:
                result = RuleResult(
                    'db',
                    result=Result.from_string(sample_info.result.name),
                    reason=sample_info.reason,
                    further_analysis=True
                )
            else:
                result = RuleResult(
                    'db',
                    result=Result.unknown,
                    reason="Datei ist dem System noch nicht bekannt",
                    further_analysis=True
                )
            session.close()
        return result

    def known(self, sample):
        """
        Check if we already know a sample with its SHA-256 hash.

        :param sample: The Sample object to check.
        """
        with self.__lock:
            is_known = False
            session = self.__Session()
            sample_info = PeekabooDatabase.__get(
                session,
                SampleInfo,
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension
            )
            if sample_info is not None:
                if sample_info.result.name != 'inProgress':
                    is_known = True
            session.close()
            return is_known

    def in_progress(self, sample):
        """
        Check if a sample is in progress using it's SHA-256 hash.

        :param sample: The Sample object to check.
        """
        with self.__lock:
            is_in_progress = False
            session = self.__Session()
            sample_info = PeekabooDatabase.__get(
                session,
                SampleInfo,
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension
            )
            if sample_info is not None:
                if sample_info.result.name == 'inProgress':
                    is_in_progress = True
            session.close()
            return is_in_progress

    def clear_in_progress(self):
        """ Remove all samples with the result 'inProgress'. """
        session = self.__Session()
        in_progress = PeekabooDatabase.__get(
            session,
            AnalysisResult,
            name='inProgress'
        )
        in_progress_samples = session.query(SampleInfo).filter_by(
            result=in_progress
        ).all()
        for in_progress_sample in in_progress_samples:
            session.query(AnalysisJournal).filter_by(
                sample=in_progress_sample
            ).delete()
        try:
            session.commit()
            logger.debug('Cleared the database from "inProgress" entries.')
        except SQLAlchemyError as e:
            session.rollback()
            raise PeekabooDatabaseError(
                'Unable to clear the database from "inProgress" entries: %s' % e
            )
        finally:
            session.close()

    def drop(self):
        """ Drop all tables of the database. """
        try:
            Base.metadata.drop_all(self.__engine)
        except SQLAlchemyError as e:
            raise PeekabooDatabaseError(
                'Unable to drop all tables of the database: %s' % e
            )

    def _init_db(self):
        """
        Initializes the Peekaboo database by creating tables and
        writing meta information to the '_meta' table.
        """
        Base.metadata.create_all(self.__engine)
        meta = PeekabooMetadata()
        meta.peekaboo_version = __version__
        meta.db_schema_version = 2
        # TODO: Get Cuckoo version.
        meta.cuckoo_version = '2.0'
        session = self.__Session()
        session.add(meta)
        '''
        session.add_all([
            AnalysisResult(name='inProgress'),
            AnalysisResult(name='unchecked'),
            AnalysisResult(name='unknown'),
            AnalysisResult(name='ignored'),
            AnalysisResult(name='checked'),
            AnalysisResult(name='good'),
            AnalysisResult(name='bad'),
        ])
        '''
        try:
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            raise PeekabooDatabaseError(
                'Cannot initialize the database: %s' % e
            )
        finally:
            session.close()

    def __exists(self, model, **kwargs):
        """
        Check whether an ORM instance exists.

        :param session: An SQLAlchemy session object.
        :param model: The model to query.
        :return: True if the ORM instance exists otherwise False.
        """
        session = self.__Session()
        instance = PeekabooDatabase.__get(session, model, **kwargs)
        session.close()
        if instance is not None:
            return True
        return False

    @staticmethod
    def __get_or_create(session, model, **kwargs):
        """
        Get an ORM instance or create it if not exist.

        :param session: An SQLAlchemy session object.
        :param model: The model to query.
        :return: A row instance.
        """
        instance = PeekabooDatabase.__get(session, model, **kwargs)
        return instance or model(**kwargs)

    @staticmethod
    def __get(session, model, **kwargs):
        """
        Get an ORM instance.

        :param session: An SQLAlchemy session object.
        :param model: The model to query.
        :return: An ORM instance or None.
        """
        return session.query(model).filter_by(**kwargs).first()

    @staticmethod
    def __create(model, **kwargs):
        """
        Create an ORM instance.

        :param model: The model to create.
        :return: An ORM instance of the given model.
        """
        return model(**kwargs)

    def __del__(self):
        self.__engine.dispose()

    def __str__(self):
        return '<PeekabooDatabase(connected="%s")>' % \
               'yes' if self.__db_con is not None else 'no'

    __repr__ = __str__



#!/usr/bin/env python

###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# setup.py                                                                    #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from setuptools import setup, find_packages
from codecs import open
from os import path, system
from sys import path as pythonpath

# Add peekaboo to PYTHONPATH
pythonpath.append(path.dirname(path.dirname(path.abspath(__file__))))
from peekaboo import __version__, __author__, __license__, __description__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '')
                    for x in all_reqs if 'git+' not in x]

# compile chown2me
system('cd bin/ && make chown2me')


setup(
    name='PeekabooAV',
    version=__version__,
    description=__description__,
    long_description=long_description,
    url='https://github.com/scVENUS/PeekabooAV.git',
    download_url='https://github.com/scVENUS/PeekabooAV/archive/master.zip',
    license=__license__,
    classifiers=[
      'Development Status :: 4 - Beta',
      'Operating System :: POSIX',
      'Programming Language :: Python',
      'Programming Language :: Python :: 2',
      'Programming Language :: Python :: 3',
      'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      'Natural Language :: English',
      'Natural Language :: German',
      'Topic :: Communications :: Email :: Filters',
    ],
    keywords='Cuckoo, Amavis',
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    author=__author__,
    install_requires=install_requires,
    dependency_links=dependency_links,
    author_email='felix.bauer@atos.net, sebastian.deiss@atos.net',
    entry_points={
        'console_scripts': [
            'peekaboo = peekaboo.daemon:run',
        ],
    }
)



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         engine.py                                                           #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import os
import logging
from shutil import copyfile
from peekaboo.ruleset import Result, RuleResult
from peekaboo.ruleset.rules import *
from peekaboo.exceptions import CuckooReportPendingException
from peekaboo.toolbox.plugins.oneanalysis import OneAnalysis
from peekaboo.toolbox.peekabooyar import contains_peekabooyar

'''
# this module contains methods and data structures which allow to
# create a ruleset to decide good or bad for any given file
#
# works together with peekaboo
# and uses cuckoo
'''


logger = logging.getLogger(__name__)


def run_analysis(sample):
    """
    function that is run by a worker for every Sample object.
    """
    process_rules(sample)
    logger.info("Rules evaluated")
    report(sample)
    one_analysis_tool = OneAnalysis()
    one_analysis_tool.queue_identical_samples(sample)  # depends on already_in_progress


def rule(sample, rule_function, args={}):
    """
    rule wrapper for in/out logging and reporting
    """
    function_name = rule_function.func_name
    logger.debug("Processing rule '%s' for %s" % (function_name, sample))

    try:
        if args:
            res = rule_function(sample, args)
        else:
            res = rule_function(sample)

        sample.add_rule_result(res)
    except CuckooReportPendingException as e:
        # in case this our Sample is requesting the Cuckoo report
        raise
    # catch all exceptions in rule
    except Exception as e:
        logger.warning("Unexpected error in '%s' for %s" % (function_name,
                                                            sample))
        logger.exception(e)
        # create "fake" RuleResult
        res = RuleResult("rule_wrapper", result=Result.unknown,
                         reason="Regel mit Fehler abgebrochen",
                         further_analysis=True)
        sample.add_rule_result(res)

    logger.info("Rule '%s' processed for %s" % (function_name, sample))
    return res


def process_rules(sample):
    s = sample
#                      ____   _   _  _      _____  ____
#                     |  _ \ | | | || |    | ____|/ ___|
#                     | |_) || | | || |    |  _|  \___ \
#                     |  _ < | |_| || |___ | |___  ___) |
#                     |_| \_\ \___/ |_____||_____||____/

# TODO (cuckooWrapper needs to check if there is other samples in pjobs with
# the same hash)
    one_analysis_tool = OneAnalysis()
    p = rule(s, one_analysis_tool.already_in_progress)
    if not p.further_analysis:
        return

    p = rule(s, known)
    if not p.further_analysis:
        return

    p = rule(s, file_larger_than, {"byte": 5})
    if not p.further_analysis:
        return

    p = rule(s, file_type_on_whitelist)
    if not p.further_analysis:
        return

    p = rule(s, file_type_on_greylist)
    if not p.further_analysis:
        return

    p = rule(s, contains_peekabooyar)
    if not p.further_analysis:
        return

    p = rule(s, office_macro)
    if not p.further_analysis:
        return

    p = rule(s, requests_evil_domain)
    if not p.further_analysis:
        return

    p = rule(s, cuckoo_evil_sig)
    if not p.further_analysis:
        return

    p = rule(s, cuckoo_score, {"higher": 4.0})
    if not p.further_analysis:
        return

    p = rule(s, cuckoo_analysis_failed)
    if not p.further_analysis:
        return

    p = rule(s, final_rule)
    if not p.further_analysis:
        return

    # active rules, non reporting
#    report(sample)

#                   __ ____   _   _  _      _____  ____
#                  / /|  _ \ | | | || |    | ____|/ ___|
#                 / / | |_) || | | || |    |  _|  \___ \
#                / /  |  _ < | |_| || |___ | |___  ___) |
#               /_/   |_| \_\ \___/ |_____||_____||____/
    return None


def report(s):
    # TODO: might be better to do this for each rule individually
    s.report()
    if s.get_result() == Result.bad:
        dump_processing_info(s)
    s.save_result()


def dump_processing_info(sample):
    """
    Saves the Cuckoo report as HTML + JSON and the meta info file (if available)
    to a directory named after the job hash.
    """
    job_hash = sample.get_job_hash()
    dump_dir = os.path.join(os.environ['HOME'], 'malware_reports', job_hash)
    if not os.path.isdir(dump_dir):
        os.makedirs(dump_dir, 0770)
    filename = sample.get_filename() + '-' + sample.sha256sum

    logger.debug('Dumping processing info to %s for sample %s' % (dump_dir, sample))

    # meta info file
    if sample.has_attr('meta_info_file'):
        try:
            copyfile(sample.get_attr('meta_info_file'),
                     os.path.join(dump_dir, filename + '.info'))
        except Exception as e:
            logger.exception(e)

    # Peekaboo's report
    try:
        with open(os.path.join(dump_dir, filename + '_report.txt'), 'w+') as f:
            f.write(sample.get_peekaboo_report())
    except Exception as e:
        logger.exception(e)

    if sample.has_attr('cuckoo_json_report_file'):
        # Cuckoo report
        try:
            # JSON
            copyfile(sample.get_attr('cuckoo_json_report_file'),
                     os.path.join(dump_dir, filename + '.json'))
        except Exception as e:
            logger.exception(e)
        try:
            # HTML
            copyfile(sample.get_attr('cuckoo_json_report_file').replace('json', 'html'),
                     os.path.join(dump_dir, filename + '.html'))
        except Exception as e:
            logger.exception(e)



#!/usr/bin/env python

###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# test.py                                                                     #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import sys
import os
import unittest


# Add Peekaboo to PYTHONPATH
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from peekaboo.sample import Sample
from peekaboo.ruleset import RuleResult, Result
from peekaboo.db import PeekabooDatabase
from peekaboo.config import _set_config


class PeekabooDummyConfig(object):
    def __init__(self):
        self.db_con = None
        self.job_hash_regex = r'/var/lib/amavis/tmp/([^/]+)/parts.*'
        self.sample_base_dir = '/tmp'
        self.chown2me_exec = 'bin/chown2me'

    def set_db_con(self, db_con):
        self.db_con = db_con

    def get_db_con(self):
        return self.db_con


class PeekabooDummyDB(object):
    def sample_info2db(self, sample):
        pass

    def sample_info_fetch(self, sha256):
        pass

    def fetch_rule_result(self, sha256):
        return RuleResult('fake_db',
                          result=Result.checked,
                          reason='Test Case',
                          further_analysis=True)

    def sample_info_update(self, sample):
        pass

    def known(self, sha256):
        return False

    def in_progress(self, sha256):
        return True

    def close(self):
        pass

    def _clear_in_progress(self):
        pass


class TestDatabase(unittest.TestCase):
    """
    Unittests for Peekaboo's database module.

    @author: Sebastian Deiss
    """
    @classmethod
    def setUpClass(cls):
        cls.test_db = os.path.abspath('./test.db')
        cls.conf = PeekabooDummyConfig()
        db_con = PeekabooDatabase('sqlite:///' + cls.test_db)
        cls.conf.set_db_con(db_con)
        _set_config(cls.conf)
        cls.sample = Sample(os.path.realpath(__file__))
        result = RuleResult('Unittest',
                            Result.unknown,
                            'This is just a test case.',
                            further_analysis=True)
        cls.sample.add_rule_result(result)
        cls.sample.determine_result()

    def test_1_analysis2db(self):
        self.conf.db_con.analysis2db(self.sample)

    def test_2_sample_info_fetch(self):
        sample_info = self.conf.db_con.sample_info_fetch(self.sample)
        self.assertEqual(self.sample.sha256sum, sample_info.sha256sum)

    def test_3_sample_info_update(self):
        result = RuleResult('Unittest',
                            Result.checked,
                            'This is another test case.',
                            further_analysis=False)
        self.sample.add_rule_result(result)
        self.sample.determine_result()
        self.conf.db_con.sample_info_update(self.sample)
        rule_result = self.conf.db_con.fetch_rule_result(self.sample)
        self.assertEqual(rule_result.result, Result.checked)
        self.assertEqual(rule_result.reason, 'This is another test case.')

    def test_4_fetch_rule_result(self):
        rule_result = self.conf.db_con.fetch_rule_result(self.sample)
        # RuleResults from the DB have 'db' as rule name
        self.assertEqual(rule_result.rule, 'db')
        self.assertEqual(rule_result.result, Result.checked)
        self.assertEqual(rule_result.reason, 'This is another test case.')
        # We assert True since the DB rule result always sets further_analysis to True
        self.assertTrue(rule_result.further_analysis)

    def test_5_known(self):
        self.assertTrue(self.conf.db_con.known(self.sample))
        self.assertFalse(self.conf.db_con.in_progress(self.sample))

    @classmethod
    def tearDownClass(cls):
        cls.conf.db_con.close()
        os.unlink(cls.test_db)


class TestSample(unittest.TestCase):
    """
    Unittests for Samples.

    @author: Sebastian Deiss
    """
    @classmethod
    def setUpClass(cls):
        cls.test_db = os.path.abspath('./test.db')
        cls.conf = PeekabooDummyConfig()
        db_con = PeekabooDatabase('sqlite:///' + cls.test_db)
        cls.conf.set_db_con(db_con)
        _set_config(cls.conf)
        cls.sample = Sample(os.path.realpath(__file__))

    def test_attribute_dict(self):
        self.sample.set_attr('Unittest', 'Hello World!')
        self.assertTrue(self.sample.has_attr('Unittest'))
        self.assertEqual(self.sample.get_attr('Unittest'), 'Hello World!')
        self.sample.set_attr('Unittest', 'Test', override=True)
        self.assertEqual(self.sample.get_attr('Unittest'), 'Test')

    def test_job_hash_regex(self):
        path_with_job_hash = '/var/lib/amavis/tmp/amavis-20170831T132736-07759-iSI0rJ4b/parts'
        sample = Sample(path_with_job_hash)
        job_hash = sample.get_job_hash()
        self.assertEqual(job_hash, 'amavis-20170831T132736-07759-iSI0rJ4b',
                         'Job hash regex is not working')
        job_hash = self.sample.get_job_hash()
        self.assertIn('peekaboo-run_analysis', job_hash)

    def test_sample_attributes(self):
        self.assertEqual(self.sample.get_filename(), 'test.py')
        self.assertEqual(self.sample.file_extension, 'py')
        self.assertTrue(self.__contains_mime(self.sample.mimetypes, 'text/x-python'))
        self.assertIsNotNone(self.sample.sha256sum)
        self.assertEqual(self.sample.job_id, -1)
        self.assertEqual(self.sample.get_result(), Result.unchecked)
        self.assertEqual(self.sample.reason,
                         'Ausschlaggebendes Ergebnis laut Datenbank: Datei ist dem System noch nicht bekannt')
        self.assertFalse(self.sample.office_macros)
        self.assertFalse(self.sample.known)

    def test_sample_attributes_with_meta_info(self):
        test_meta_info = '[attachment]\n'
        test_meta_info += 'full_name     : /tmp/test.pyc\n'
        test_meta_info += 'name_declared : test.pyc\n'
        test_meta_info += 'type_declared : application/x-bytecode.python\n'
        test_meta_info += 'type_long     : application/x-python-bytecode\n'
        test_meta_info += 'type_short    : pyc\n'
        test_meta_info += 'size          : 200\n'
        test_meta_info += 'digest        :\n'
        test_meta_info += 'attributes    :\n'
        test_meta_info += 'queue_id      :\n'
        with open('./test_meta_info.info', 'w+') as f:
            f.write(test_meta_info)
        self.sample.load_meta_info('./test_meta_info.info')
        self.assertEqual(self.sample.file_extension, 'pyc')

    def test_sample_without_suffix(self):
        test_meta_info = '[attachment]\n'
        test_meta_info += 'full_name     : /tmp/junk\n'
        test_meta_info += 'name_declared : Report.docx\n'
        test_meta_info += 'type_declared : application/vnd.openxmlformats-officedocument.wordprocessingml.document\n'
        test_meta_info += 'type_long     : application/vnd.openxmlformats-officedocument.wordprocessingml.document\n'
        test_meta_info += 'type_short    : docx\n'
        test_meta_info += 'size          : 212\n'
        test_meta_info += 'digest        :\n'
        test_meta_info += 'attributes    :\n'
        test_meta_info += 'queue_id      :\n'
        with open('./junk.info', 'w+') as f:
            f.write(test_meta_info)
        sample = Sample('junk')
        self.assertEqual(sample.file_extension, '')
        sample.load_meta_info('./junk.info')
        self.assertEqual(sample.file_extension, 'docx')

    @classmethod
    def tearDownClass(cls):
        cls.conf.db_con.close()
        os.unlink(cls.test_db)
        os.unlink('./test_meta_info.info')
        os.unlink('./junk.info')

    def __contains_mime(self, mimetypes, mime):
        if mime in mimetypes:
            return True
        return False


def main():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestSample))
    suite.addTest(unittest.makeSuite(TestDatabase))
    # TODO: We need more tests!!!

    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(suite)

    if not result.wasSuccessful():
        sys.exit(1)


if __name__ == '__main__':
    main()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         rules.py                                                            #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2018  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import traceback
import re
import logging
from peekaboo.config import get_config
from peekaboo.ruleset import Result, RuleResult


logger = logging.getLogger(__name__)


def known(s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    db = get_config().get_db_con()
    if db.known(s):
        sample_info = db.sample_info_fetch(s)
        return RuleResult(position,
                          result=sample_info.get_result(),
                          reason=sample_info.reason,
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Datei ist dem System noch nicht bekannt",
                      further_analysis=True)


def file_larger_than(s, args):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    size = args['byte']
    if s.file_size > size:
        return RuleResult(position,
                          result=Result.unknown,
                          reason="Datei hat mehr als %d bytes"
                          % size,
                          further_analysis=True)

    return RuleResult(position,
                      result=Result.ignored,
                      reason="Datei ist nur %d bytes lang"
                      % s.file_size,
                      further_analysis=False)


def file_type_on_whitelist(s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    # MIME file types that should not be analyzed
    whitelist = [
        # MIME types
        'text/plain',
        'text/html',
        'message/rfc822',
        'None',
        # magic
        'inode/x-empty',
        'text/plain',
        'application/pkcs7-signature',
    ]

    # analysis wanted for file type
    mtypes = s.mimetypes
    n = 0
    logger.debug("Filetype is %s" % mtypes)
    for mtype in mtypes:
        if mtype in whitelist:
            n = n + 1

    # TODO: Check if one hit on the whitelist is sufficient!
    if n != len(mtypes):
        logger.debug('Length of n: %s, Content of mtypes: %s' % (str(n),
                                                                 str(mtypes)))

    if n >= len(mtypes):
        return RuleResult(position,
                          result=Result.ignored,
                          reason="Dateityp ist auf Whitelist",
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Dateityp ist nicht auf Whitelist",
                      further_analysis=True)


def file_type_on_greylist(s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    # MIME file types that should be analyzed
    greylist = [
        # magic
        'application/octet-stream',
        'application/vnd.ms-excel',
        'application/pdf',
        'application/javascript',
        'application/pdf',
        'application/vnd.ms-excel',
        'application/vnd.ms-excel.sheet.macroEnabled.12',
        'application/vnd.ms-word.document.macroEnabled.12',
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        'application/x-7z-compressed',
        'application/x-ms-dos-executable',
        'application/x-dosexec',
        'application/x-vbscript',
        'application/zip',
        'application/x-rar',
        'application/msword',
        # 'message/rfc822',
        # 'text/html',
        'text/x-msdos-batch',
        # MIME types
        'text/x-sh',
        'text/x-python',
        'image/png',
        'image/jpeg',
        'application/zip',
        'application/x-silverlight',
        'application/x-python-code',
        'application/x-msdos-program',
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        'application/vnd.openxmlformats-officedocument.presentationml.presentation',
        'application/vnd.oasis.opendocument.text',
        'application/vnd.oasis.opendocument.spreadsheet',
        'application/vnd.oasis.opendocument.presentation',
        'application/vnd.ms-word.template.macroEnabled.12',
        'application/vnd.ms-powerpoint',
        'application/vnd.ms-excel.template.macroEnabled.12',
        'application/vnd.ms-excel',
        'application/pdf',
        'application/msword'
    ]

    mtypes = s.mimetypes
    logger.debug("filetype is %s" % mtypes)

    for mtype in mtypes:
        if mtype in greylist:
            return RuleResult(position,
                              result=Result.unknown,
                              reason="Dateityp ist auf der Liste der zu analysiserenden Typen",
                              further_analysis=True)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Dateityp ist nicht auf der Liste der zu analysierenden Typen (%s)"
                      % (str(s.mimetypes)),
                      further_analysis=False)


def cuckoo_evil_sig(s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    # signatures that if matched mark a sample as bad
    # list all installed signatures
    # grep -o "description.*" -R . ~/cuckoo2.0/modules/signatures/
    bad_sigs = [
        "A potential heapspray has been detected. .*",
        "A process attempted to delay the analysis task.",
        "Attempts to detect Cuckoo Sandbox through the presence of a file",
        "Attempts to modify desktop wallpaper",
        "Checks amount of memory in system, this can be used to detect " +
        "virtual machines that have a low amount of memory available",
        "Checks the version of Bios, possibly for anti-virtualization",
        "Collects information on the system (ipconfig, netstat, systeminfo)",
        "Connects to an IRC server, possibly part of a botnet",
        "Connects to Tor Hidden Services through Tor2Web",
        "Creates a suspicious process",
        "Creates a windows hook that monitors keyboard input (keylogger)",
        "Creates executable files on the filesystem",
        "Creates known Upatre files, registry keys and/or mutexes",
        # "Creates (office) documents on the filesystem",
        "Detects the presence of Wine emulator",
        "Detects VirtualBox through the presence of a file",
        "Detects VirtualBox through the presence of a registry key",
        "Detects VirtualBox through the presence of a window",
        "Detects VirtualBox using WNetGetProviderName trick",
        "Detects VMWare through the in instruction feature",
        "Detects VMWare through the presence of a registry key",
        "Detects VMWare through the presence of various files",
        "Executes javascript",
        "Executes one or more WMI queries",
        "File has been identified by .* AntiVirus engines on VirusTotal as " +
        "malicious",
        "Installs itself for autorun at Windows startup",
        "Looks for known filepaths where sandboxes execute samples",
        "Looks for the Windows Idle Time to determine the uptime",
        "Makes SMTP requests, possibly sending spam",
        "This sample modifies more than .* files through suspicious ways,",
        "Network communications indicative of a potential document or script" +
        " payload download was initiated by the process wscript.exe",
        "One of the processes launched crashes",
        "One or more of the buffers contains an embedded PE file",
        "One or more potentially interesting buffers were extracted, these " +
        "generally",
        "Potentially malicious URL found in document",
        "Queries for the computername",
        "Queries the disk size.*",
        "Raised Suricata alerts",
        "Starts servers listening on {0}",
        "Steals private information from local Internet browsers",
        "Suspicious Javascript actions",
        "Tries to detect analysis programs from within the browser",
        "Tries to locate whether any sniffers are installed",
        "Wscript.exe initiated network communications indicative of a script" +
        " based payload download",
    ]

    sigs = []

    # look through matched signatures
    for descr in s.cuckoo_report.signatures:
        logger.debug(descr['description'])
        sigs.append(descr['description'])

    # check if there is a "bad" signatures and return bad
    matched_bad_sigs = []
    for sig in bad_sigs:
        match = re.search(sig, str(sigs))
        if match:
            matched_bad_sigs.append(sig)

    if len(matched_bad_sigs) == 0:
        return RuleResult(position,
                          result=Result.unknown,
                          reason="Keine Signatur erkannt die auf Schadcode hindeutet",
                          further_analysis=True)

    return RuleResult(position,
                      result=Result.bad,
                      reason="Folgende Signaturen wurden erkannt: %s"
                      % ''.ljust(8).join(["%s\n" % s for s in matched_bad_sigs]),
                      further_analysis=False)


def cuckoo_score(s, args):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    threshold = args['higher']
    if s.cuckoo_report.score >= threshold:
        return RuleResult(position,
                          result=Result.bad,
                          reason="Cuckoo score >= %s: %s"
                          % (threshold, s.cuckoo_report.score),
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Cuckoo score < %s: %s"
                      % (threshold, s.cuckoo_report.score),
                      further_analysis=True)


def office_macro(s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    if s.office_macros:
        return RuleResult(position,
                          result=Result.bad,
                          reason="Die Datei beinhaltet ein Office-Makro",
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Die Datei beinhaltet kein erkennbares Office-Makro",
                      further_analysis=True)


def requests_evil_domain(s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    evil_domains = ["canarytokens.com"]

    for d in s.requested_domains:
        if d in evil_domains:
            return RuleResult(position,
                              result=Result.bad,
                              reason="Die Datei versucht mindestens eine Domain aus der Blacklist zu kontaktieren (%s)" % d,
                              further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Datei scheint keine Domains aus der Blacklist kontaktieren zu wollen",
                      further_analysis=True)


def cuckoo_analysis_failed(s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    if s.cuckoo_analysis_failed:
        return RuleResult(position,
                          result=Result.bad,
                          reason="Die Verhaltensanalyse durch Cuckoo hat einen Fehler Produziert und konnte nicht erfolgreich abgeschlossen werden",
                          further_analysis=False)

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Die Verhaltensanalyse durch Cuckoo wurde erfolgreich abgeschlossen",
                      further_analysis=True)


def final_rule(s):
    tb = traceback.extract_stack()
    tb = tb[-1]
    position = "%s:%s" % (tb[2], tb[1])

    return RuleResult(position,
                      result=Result.unknown,
                      reason="Datei scheint keine erkennbaren Schadroutinen zu starten",
                      further_analysis=True)
